/*
 * Marble.java
 *
 * Created on 19. heinkuuta 2004, 15:27
 */

package marbles;

import graphlib.*;

/**
 * Rigid body Marble.
 * @author  millat
 */
public class Marble {
  
  public static final int STATE_SIZE = 18;
  public static final int VECTOR_SIZE = 3;
  public static final int QUATERNION_SIZE = 4;
  
  // private static final MatrixCalculations mc = new MatrixCalculations();
  
  /** Constant quantities */
  private double mass           = 1;
  private double radius         = 1;
  private double [][] IBody     = null;
  private double [][] IBodyInv  = null;
  
  /** State variables */
  private double [] x           = null;     // Massakeskipisteen sijainti
  private double [] q           = null;     // Kvaternio
  private double [] P           = null;     // Kokonaisliikemr
  private double [] L           = null;     // Kulmaliikemr
  
  /** Derived quantities (auxiliary variables) */
  private double [][] IInv      = null;     // Inertiatensorin knteinen
  private double [][] R         = null;     // Rotation matrix
  private double [] v           = null;     // Linear velocity
  private double [] omega       = null;     // Angular velocity
  
  /** Computed quantities */
  double [] force               = null;     // Ulkoinen kokonaisvoima
  double [] torque              = null;     // Ulkoisen voiman aiheuttama vntmomentti
  
  
  /** Creates a new instance of Marble */
  public Marble() {
    x = new double [] { 0, 0, 0 };
  }
  
  /** Creates a new instance of Marble */
  public Marble(double [] x, double radius, double mass) {
    this(x, new double [] { 0, 0, 0 }, radius, mass);
  }
  
  /** Creates a new instance of Marble */
  public Marble(double [] x, double [] v, double radius, double mass) {
    if (x.length == 3)
      this.x = x;
    if (radius > 0)
      this.radius = radius;
    if (mass > 0)
      this.mass = mass;
    this.v = v;
    IBody = MatrixCalculations.multiplyMatrixByX(MatrixCalculations.unitMatrix() ,(2*mass*radius*radius)/5.0);
    IBodyInv = MatrixCalculations.multiplyMatrixByX(MatrixCalculations.unitMatrix(), 5.0/(2*mass*radius*radius));
    P = MatrixCalculations.initVector(VECTOR_SIZE);
    L = MatrixCalculations.initVector(VECTOR_SIZE);
    //L[1] = 40.5;
    q = new double[] { 1.0, 0.0, 0.0, 0.0}; //pitisi vastata 0:n kiertoa
    // q = MatrixCalculations.initVector(QUATERNION_SIZE);
    force = MatrixCalculations.initVector(VECTOR_SIZE);
    torque = MatrixCalculations.initVector(VECTOR_SIZE);
  }
  
  
  /** Get location (drawing) */
  public double getX(int index)  { //throws Exception
    //      if (index < 0 || index > 2)
    //         throw new Exception("Bad index!");
    return x[index];
  }
  
  public double[] getX()  {
    return x;
  }
  
  /** Get radius (drawing) */
  public double getRadius() {
    return radius;
  }
  
  //// kontaktia varten listtyj
/*  public double[] getForce() {
    return force;
  }
  
  public double[] getTorque() {
    return torque;
  }
 */
  public double[] getOmega() {
    return omega;
  }
 /* 
  public double[] getL() {
    return L;
  }
  
  */
  public double[][] getIinv() {
    return IInv;
  }
  
  /** Get Q (draw rotation) */
  public double [] getQ4DVector() throws MatrixException {
    return MatrixCalculations.quaternionToMatrix4DVector(q);
  }
  
  /** Get mass (forces) */
  public double getMass() {
    return mass;
  }
  
  public double[] getAcceleration() throws MatrixException {
    return MatrixCalculations.multiplyVectorByX(force, 1.0/mass);
  }
  
  public double[] getAngularAcceleration() throws MatrixException {
      return MatrixCalculations.multiplyMatrixByVector(IInv, MatrixCalculations.sumVectors(torque, MatrixCalculations.crossProduction(L, omega)));
  }
  
  public double[] getPointVelocity(double [] point) throws MatrixException {
    double [] a = MatrixCalculations.subtractionOfVectors(point,x);
    return MatrixCalculations.sumVectors(v, MatrixCalculations.crossProduction(omega, a));
  }
  
  public void clearForce() throws MatrixException {
    this.force = MatrixCalculations.initVector(3);
  }
  
  public void addForce(double [] f) throws MatrixException {
    this.force = MatrixCalculations.sumVectors(this.force, f);
  }
  
  /*public void addL(double [] J, double [] r) throws MatrixException {
    L = MatrixCalculations.sumVectors(L, MatrixCalculations.crossProduction(r, J));
    omega = MatrixCalculations.multiplyMatrixByVector(IInv, L);
  }*/
  
   public void setL(double [] J, double [] r) throws MatrixException {
    L = J;
    omega = MatrixCalculations.multiplyMatrixByVector(IInv, L);
  }

  public double[] getV() {
    return v;
  }
  
  public void addImpulse(double [] J, double [] r) throws MatrixException {
    P = MatrixCalculations.sumVectors(P, J);
    L = MatrixCalculations.sumVectors(L, MatrixCalculations.crossProduction(r, J));
    // Laskennalliset nist riippuvat arvot
    v = MatrixCalculations.multiplyVectorByX(P, 1 / mass);
    omega = MatrixCalculations.multiplyMatrixByVector(IInv, L);
  }
  
  public void addContactForce(double[] f, double[] pa) throws MatrixException {
   //  clearForce();
      addForce(f);
    //  torque = MatrixCalculations.sumVectors(torque, MatrixCalculations.crossProduction(pa, f));
  }
  
  /** Return current state as an array */
  public double [] stateToArray() {
    double [] y = new double [STATE_SIZE];
    int yIndex = 0;
    // Copy state variables - Kopioidaan tilamuuttujat
    for (int i = 0; i < VECTOR_SIZE; i++)
      y[yIndex++] = x[i];
    for (int i = 0; i < QUATERNION_SIZE; i++)
      y[yIndex++] = q[i];
    for (int i = 0; i < VECTOR_SIZE; i++)
      y[yIndex++] = P[i];
    for (int i = 0; i < VECTOR_SIZE; i++)
      y[yIndex++] = L[i];
    
    return y;
  }
  
  /** Derivates of state variables to array */
  public double [] ddtStateToArray() throws MatrixException {
    double [] y = new double [STATE_SIZE];
    int yIndex = 0;
    for(int i=0; i<VECTOR_SIZE; i++)
      y[yIndex++] = v[i];
    double [] tmp = new double [] { 0, omega[0], omega[1], omega[2] };
    double [] quaternion = MatrixCalculations.multiplyVectorByX(
    MatrixCalculations.quaternionProduct(tmp, q), 0.5);
    for(int i=0; i< QUATERNION_SIZE; i++)
      y[yIndex++] = quaternion[i];
    
    for(int i=0; i< VECTOR_SIZE; i++)
      y[yIndex++] = force[i];
    for(int i=0; i< VECTOR_SIZE; i++)
      y[yIndex++] = torque[i];
    
    return y;
  }
  
  /** Update state from array */
  public void arrayToState(double [] y) throws MatrixException {
    int yIndex = 0;
    // Copy state variables - Kopioidaan tilamuuttujat
    for (int i = 0; i < VECTOR_SIZE; i++)
      x[i] = y[yIndex++];
    for (int i = 0; i < QUATERNION_SIZE; i++)
      q[i] = y[yIndex++];
    for (int i = 0; i < VECTOR_SIZE; i++)
      P[i] = y[yIndex++];
    for (int i = 0; i < VECTOR_SIZE; i++)
      L[i] = y[yIndex++];
    
    // Compute auxiliary values - Lasketaan johdannaiset
    v = MatrixCalculations.multiplyVectorByX(P, 1 / mass);
    // Update R with normalized quaternion
    R = MatrixCalculations.quaternionToMatrix(MatrixCalculations.normalize(q));
    IInv = MatrixCalculations.multiplyMatrixByMatrix(R, MatrixCalculations.multiplyMatrixByMatrix(IBodyInv, MatrixCalculations.transpose(R)));
    omega = MatrixCalculations.multiplyMatrixByVector(IInv, L);
  }
  
  
}
