/*
 * MarbleCanvas.java
 *
 * Created on 19. heinkuuta 2004, 15:33
 */

package marbles;

import marbles.log.*;
import graphlib.*;

import gl4java.*;
import gl4java.awt.*;
import gl4java.utils.glut.*;

/**
 * Canvas that draws the plane and marbles in their current positions when display() is called.
 * @author  millat
 */
public class MarbleCanvas extends GLCanvas implements Display {
  
  private static final double CAMERA_MOVE_STEP = 1;
  private static final double CAMERA_TURN_ANGLE = 5;
  
  private GLUTFunc glut = null;
  private double [] cameraPoint = new double [] { -10, -10, 2 };
  
  private MarbleContainer marbles = null;
  private LogInterface log = null;
  
  /** Creates a new instance of MarbleCanvas */
  public MarbleCanvas(int width, int height, MarbleContainer marbles, LogInterface log) {
    super(width, height);
    GLContext.gljNativeDebug = false;
    GLContext.gljClassDebug = false;
    this.marbles = marbles;
    this.log = log;
    marbles.setDisplay(this);
  }
  
  public void preInit() {
    doubleBuffer = true;
    stereoView = false;
  }
  
  public void init() {
    glut = new GLUTFuncLightImpl(gl, glu);
    reshape(getSize().width, getSize().height);
    
    gl.glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    gl.glShadeModel(GL_FLAT);
  }
  
  private void drawXYPlane(double [] downleft, double [] upright, double z, double gridsize) {
    gl.glBegin(GL_LINES);
    
    // Horizontal lines
    double x1 = downleft[0];
    double y = downleft[1];
    double x2 = upright[0];
    
    while (y <= upright[1]) {
      gl.glVertex3d(x1, y, z);
      gl.glVertex3d(x2, y, z);
      y += gridsize;
    }
    
    // Vertical lines
    double y1 = downleft[1];
    double x = downleft[0];
    double y2 = upright[1];
    
    while (x <= upright[0]) {
      gl.glVertex3d(x, y1, z);
      gl.glVertex3d(x, y2, z);
      x += gridsize;
    }
    
    gl.glEnd();
  }
  
  public void display() {
    if (glj.gljMakeCurrent() == false) return;
    gl.glClear(GL_COLOR_BUFFER_BIT);
    
    gl.glColor3f(0.5f, 0.5f, 0.5f);
    drawXYPlane(new double [] { -10, -10 }, new double [] { 10, 10 }, 0, 1);
    
    gl.glColor3f(1.0f, 1.0f, 1.0f);
    
    gl.glBegin(GL_POINTS);
    gl.glVertex3f(0, 0, 0);
    gl.glEnd();
    
    for (int i = 0; i < marbles.getMarbleCount(); i++) {
      try {
        Marble tmp = marbles.getMarble(i);
        gl.glPushMatrix();
        double [] q = tmp.getQ4DVector();
        gl.glTranslated(tmp.getX(0), tmp.getX(1), tmp.getX(2));
        gl.glMultMatrixd(q);
        glut.glutWireSphere(tmp.getRadius(), 10, 8);
        gl.glPopMatrix();
      } catch (Exception exp) {
        log.error(exp);
      }
    }
    
    glj.gljSwap();
    glj.gljFree();
  }
  
  public void reshape(int w, int h) {
    gl.glViewport(0, 0, w, h);
    gl.glMatrixMode(GL_PROJECTION);
    gl.glLoadIdentity();
    glu.gluPerspective(60,  w/ h, 1, 20);
    
    gl.glMatrixMode(GL_MODELVIEW);
    gl.glLoadIdentity();
    glu.gluLookAt(cameraPoint[0], cameraPoint[1], cameraPoint[2], 0, 0, 0, 0, 0, 1);
  }
  
  private void refreshCamera() {
    //log.info(cameraPoint[0] + ", " + cameraPoint[1] + ", " + cameraPoint[2]);
    glj.gljMakeCurrent();
    this.reshape(getSize().width, getSize().height);
    if (!marbles.isAnimationRunning()) {
      this.display();
    }
  }
  
  public void cameraUp() {
    cameraPoint[2] += CAMERA_MOVE_STEP;
    refreshCamera();
  }
  
  public void cameraDown() {
    if (cameraPoint[2] >= CAMERA_MOVE_STEP) {
      cameraPoint[2] -= CAMERA_MOVE_STEP;
      refreshCamera();
    }
  }
  
  public void cameraLeft() {
    try {
      cameraPoint = MatrixCalculations.multiplyMatrixByVector(MatrixCalculations.R(-CAMERA_TURN_ANGLE), cameraPoint);
    } catch (Exception exp) {
      log.error(exp);
    }
    refreshCamera();
  }
  
  public void cameraRight() {
    try {
      cameraPoint = MatrixCalculations.multiplyMatrixByVector(MatrixCalculations.R(CAMERA_TURN_ANGLE), cameraPoint);
      refreshCamera();
    } catch (Exception exp) {
      log.error(exp);
    }
  }
  
  public void zoomIn() {
    try {
      if (MatrixCalculations.length(cameraPoint) > 1) {
        cameraPoint = MatrixCalculations.sumVectors(cameraPoint, MatrixCalculations.multiplyVectorByX(cameraPoint, -1 / MatrixCalculations.length(cameraPoint)));
        refreshCamera();
      }
    } catch (Exception exp) {
      log.error(exp);
    }
  }
  
  public void zoomOut() {
    try {
      cameraPoint = MatrixCalculations.sumVectors(cameraPoint, MatrixCalculations.multiplyVectorByX(cameraPoint, 1 / MatrixCalculations.length(cameraPoint)));
      refreshCamera();
    } catch (Exception exp) {
      log.error(exp);
    }
  }
  
}
